# -*- coding: utf-8 -*-
"""
@author: Sam Schott  (ss2151@cam.ac.uk)

(c) Sam Schott; This work is licensed under the MIT licence.

This module contains the default configuration and state values and functions to return
existing config or state instances for a specified config_name.

"""

import copy
import logging
import threading

from .base import get_conf_path, get_data_path
from .user import UserConfig

logger = logging.getLogger(__name__)
CONFIG_DIR_NAME = 'maestral'

# =============================================================================
#  Defaults
# =============================================================================

DEFAULTS = [
    ('main',
     {
         'path': '',  # dropbox folder location
         'default_dir_name': 'Dropbox ({})',  # default dropbox folder name
         'excluded_items': [],  # files and folders excluded from sync
     }
     ),
    ('account',
     {
         'account_id': '',  # dropbox account id, must match the saved account key
     }
     ),
    ('app',
     {
         'notification_level': 15,  # desktop notification level, default to FILECHANGE
         'log_level': 20,  # log level for journal and file, default to INFO
         'update_notification_interval': 60 * 60 * 24 * 7,  # default to weekly
         'analytics': False,  # automatic errors reports with bugsnag, default to disabled
         'keyring': 'automatic',  # keychain backend to use for credential storage
     }
     ),
    ('sync',
     {
         'reindex_interval': 60 * 60 * 24 * 7,  # default to weekly
         'max_cpu_percent': 20.0,  # max usage target per cpu core, default to 20%
     }
     )
]

DEFAULTS_STATE = [
    ('account',  # account state, periodically updated from dropbox servers
     {
         'email': '',
         'display_name': '',
         'abbreviated_name': '',
         'type': '',
         'usage': '',
         'usage_type': '',
     }
     ),
    ('app',  # app state
     {
         'update_notification_last': 0.0,
         'latest_release': '0.0.0',
     }
     ),
    ('sync',  # sync state, updated by monitor
     {
         'cursor': '',  # remote cursor: represents last state synced from dropbox
         'lastsync': 0.0,  # local cursor: time-stamp of last upload
         'last_reindex': 0.0,  # time-stamp of full last reindexing
         'download_errors': [],  # failed downloads to retry on next sync
         'pending_downloads': [],  # incomplete downloads to retry on next sync
         'recent_changes': [],  # cached list of recent changes to display in GUI / CLI
     }
     ),
]

# IMPORTANT NOTES:
# 1. If you want to *change* the default value of a current option, you need to
#    do a MINOR update in config version, e.g. from 3.0.0 to 3.1.0
# 2. If you want to *remove* options that are no longer needed in our codebase,
#    or if you want to *rename* options, then you need to do a MAJOR update in
#    version, e.g. from 3.0.0 to 4.0.0
# 3. You don't need to touch this value if you're just adding a new option
CONF_VERSION = '12.0.0'


# =============================================================================
# Factories
# =============================================================================

_config_instances = dict()
_state_instances = dict()

_config_lock = threading.Lock()
_state_lock = threading.Lock()


def MaestralConfig(config_name):
    """
    Returns existing config instance or creates a new one.

    :param str config_name: Name of maestral configuration to run. A new config file will
        be created if none exists for the given config_name.
    :return: Maestral config instance which saves any changes to the drive.
    :rtype: UserConfig
    """

    global _config_instances

    with _config_lock:

        try:
            return _config_instances[config_name]
        except KeyError:
            defaults = copy.deepcopy(DEFAULTS)
            # set default dir name according to config
            for sec, options in defaults:
                if sec == 'main':
                    options['default_dir_name'] = f'Dropbox ({config_name.title()})'

            config_path = get_conf_path(CONFIG_DIR_NAME)

            try:
                conf = UserConfig(
                    config_path, config_name, defaults=defaults, version=CONF_VERSION,
                    load=True, backup=True, raw_mode=True, remove_obsolete=True
                )
            except OSError:
                conf = UserConfig(
                    config_path, config_name, defaults=defaults, version=CONF_VERSION,
                    load=False, backup=True, raw_mode=True, remove_obsolete=True
                )

            _config_instances[config_name] = conf
            return conf


def MaestralState(config_name):
    """
    Returns existing state instance or creates a new one.

    :param str config_name: Name of maestral configuration to run. A new state file will
        be created if none exists for the given config_name.
    :return: Maestral state instance which saves any changes to the drive.
    :rtype: UserConfig
    """

    global _state_instances

    with _state_lock:

        try:
            return _state_instances[config_name]
        except KeyError:
            state_path = get_data_path(CONFIG_DIR_NAME)

            try:
                state = UserConfig(
                    state_path, config_name, defaults=DEFAULTS_STATE,
                    version=CONF_VERSION, load=True, backup=True, raw_mode=True,
                    remove_obsolete=True, suffix='.state'
                )
            except OSError:
                state = UserConfig(
                    state_path, config_name, defaults=DEFAULTS_STATE,
                    version=CONF_VERSION, load=False, backup=True, raw_mode=True,
                    remove_obsolete=True, suffix='.state'
                )

            _state_instances[config_name] = state
            return state



# -*- coding: utf-8 -*-
"""
@author: Sam Schott  (ss2151@cam.ac.uk)

(c) Sam Schott; This work is licensed under the MIT licence.

This module defines functions to start and stop the sync daemon and retrieve proxy objects
for a running daemon.

"""

# system imports
import sys
import os
import time
import signal
import traceback
import enum
import subprocess
from shlex import quote
import threading
import fcntl
import struct
import tempfile

# external imports
import Pyro5.errors
from Pyro5.api import Daemon, Proxy, expose, oneway
from Pyro5.serializers import SerpentSerializer
from fasteners import InterProcessLock

# local imports
from maestral.errors import SYNC_ERRORS, FATAL_ERRORS
from maestral.constants import IS_FROZEN, IS_MACOS
from maestral.utils.appdirs import get_runtime_path


threads = dict()
URI = 'PYRO:maestral.{0}@{1}'


def freeze_support():
    """Freeze support for multiprocessing and daemon startup. This works by checking for
    '--multiprocessing-fork' and '--frozen-daemon' command line arguments. Call this
    function at the entry point of the executable, as soon as possible and ideally before
    any heavy imports."""

    import argparse
    import multiprocessing as mp

    mp.freeze_support()

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-c', '--config-name', default='maestral')
    parser.add_argument('--frozen-daemon', action='store_true')
    parsed_args, remaining = parser.parse_known_args()

    if parsed_args.frozen_daemon:
        start_maestral_daemon(parsed_args.config_name)
        sys.exit()


class Exit(enum.Enum):
    """Enumeration of daemon exit results."""
    Ok = 0
    Killed = 1
    NotRunning = 2
    Failed = 3


class Start(enum.Enum):
    """Enumeration of daemon start results."""
    Ok = 0
    AlreadyRunning = 1
    Failed = 2


# ==== error serialization ===============================================================

def serpent_deserialize_api_error(class_name, d):
    """
    Deserializes a :class:`errors.MaestralApiError`.

    :param str class_name: Name of class to deserialize.
    :param dict d: Dictionary of serialized class.
    :returns: Class instance.
    :rtype: :class:`errors.MaestralApiError`
    """
    # import maestral errors for evaluation
    import maestral.errors  # noqa: F401

    cls = eval(class_name)
    err = cls(*d['args'])
    for a_name, a_value in d['attributes'].items():
        setattr(err, a_name, a_value)

    return err


for err_cls in list(SYNC_ERRORS) + list(FATAL_ERRORS):
    SerpentSerializer.register_dict_to_class(
        err_cls.__module__ + '.' + err_cls.__name__,
        serpent_deserialize_api_error
    )


# ==== interprocess locking ==============================================================

def _get_lockdata():

    try:
        os.O_LARGEFILE
    except AttributeError:
        start_len = 'll'
    else:
        start_len = 'qq'

    if (sys.platform.startswith(('netbsd', 'freebsd', 'openbsd'))
            or sys.platform == 'darwin'):
        if struct.calcsize('l') == 8:
            off_t = 'l'
            pid_t = 'i'
        else:
            off_t = 'lxxxx'
            pid_t = 'l'

        fmt = off_t + off_t + pid_t + 'hh'
        pid_index = 2
        lockdata = struct.pack(fmt, 0, 0, 0, fcntl.F_WRLCK, 0)
    # elif sys.platform.startswith('gnukfreebsd'):
    #     fmt = 'qqihhi'
    #     pid_index = 2
    #     lockdata = struct.pack(fmt, 0, 0, 0, fcntl.F_WRLCK, 0, 0)
    # elif sys.platform in ('hp-uxB', 'unixware7'):
    #     fmt = 'hhlllii'
    #     pid_index = 2
    #     lockdata = struct.pack(fmt, fcntl.F_WRLCK, 0, 0, 0, 0, 0, 0)
    elif sys.platform.startswith('linux'):
        fmt = 'hh' + start_len + 'ih'
        pid_index = 4
        lockdata = struct.pack(fmt, fcntl.F_WRLCK, 0, 0, 0, 0, 0)
    else:
        raise RuntimeError(f'Unsupported platform {sys.platform}')

    return lockdata, fmt, pid_index


class Lock:
    """
    A inter-process and inter-thread lock. This reuses uses code from oslo.concurrency
    but provides non-blocking acquire. Use the :meth:`singleton` class method to retrieve
    an existing instance for thread-safe usage.
    """

    _instances = dict()
    _singleton_lock = threading.Lock()

    @classmethod
    def singleton(cls, name, lock_path=None):
        """
        Retrieve an existing lock object with a given 'name' or create a new one. Use this
        method for thread-safe locks.

        :param str name: Name of lock file.
        :param str lock_path: Directory for lock files. Defaults to the temporary
            directory returned by :func:`tempfile.gettempdir()` if not given.
        """

        with cls._singleton_lock:
            try:
                instance = cls._instances[name]
            except KeyError:
                instance = cls(name, lock_path)
                cls._instances[name] = instance

            return instance

    def __init__(self, name, lock_path=None):

        self.name = name
        dirname = lock_path or tempfile.gettempdir()
        lock_path = os.path.join(dirname, name)

        self._internal_lock = threading.Semaphore()
        self._external_lock = InterProcessLock(lock_path)

        self._lock = threading.RLock()

    def acquire(self):
        """
        Attempts to acquire the given lock.

        :returns: Whether or not the acquisition succeeded.
        :rtype: bool
        """

        with self._lock:
            locked_internal = self._internal_lock.acquire(blocking=False)

            if not locked_internal:
                return False

            try:
                locked_external = self._external_lock.acquire(blocking=False)
            except Exception:
                self._internal_lock.release()
                raise
            else:

                if locked_external:
                    return True
                else:
                    self._internal_lock.release()
                    return False

    def release(self):
        """Release the previously acquired lock."""
        with self._lock:
            self._external_lock.release()
            self._internal_lock.release()

    def locked(self):
        """Checks if the lock is currently held by any thread or process."""
        with self._lock:
            gotten = self.acquire()
            if gotten:
                self.release()
            return not gotten

    def locking_pid(self):
        """
        Returns the PID of the process which currently holds the lock or ``None``. This
        should work on macOS, OpenBSD and Linux but may fail on some platforms. Always use
        :meth:`locked` to check if the lock is held by any process.

        :returns: The PID of the process which currently holds the lock or ``None``.
        :rtype: int
        """

        with self._lock:

            if self._external_lock.acquired:
                return os.getpid()

            try:
                # don't close again in case we are the locking process
                self._external_lock._do_open()
                lockdata, fmt, pid_index = _get_lockdata()
                lockdata = fcntl.fcntl(self._external_lock.lockfile,
                                       fcntl.F_GETLK, lockdata)

                lockdata_list = struct.unpack(fmt, lockdata)
                pid = lockdata_list[pid_index]

                if pid > 0:
                    return pid

            except OSError:
                pass


# ==== helpers for daemon management =====================================================

def _escape_spaces(string):
    return string.replace(' ', '_')


def _sigterm_handler(signal_number, frame):
    sys.exit()


def _send_term(pid):
    try:
        os.kill(pid, signal.SIGTERM)
    except ProcessLookupError:
        pass


class MaestralLock:
    """
    An inter-process and inter-thread lock for Maestral. This is a wrapper around
    :class:`Lock` which fills out the appropriate lockfile name and directory for the
    given config name.

    :param str config_name: The name of the Maestral configuration.
    """

    def __new__(cls, config_name):
        name = f'{config_name}.lock'
        path = get_runtime_path('maestral')
        return Lock.singleton(name, path)


def sockpath_for_config(config_name):
    """
    Returns the unix socket location to be used for the config. This should default to
    the apps runtime directory + '/maestral/CONFIG_NAME.sock'.
    """
    return get_runtime_path('maestral', f'{config_name}.sock')


def lockpath_for_config(config_name):
    return get_runtime_path('maestral', f'{config_name}.lock')


def get_maestral_pid(config_name):
    """
    Returns Maestral's PID if the daemon is running, ``None`` otherwise.

    :param str config_name: The name of the Maestral configuration.
    :returns: The daemon's PID.
    :rtype: int
    """

    return MaestralLock(config_name).locking_pid()


def is_running(config_name):
    """
    Checks if a daemon is currently running.

    :param str config_name: The name of the Maestral configuration.
    :returns: Whether the daemon is running.
    :rtype: bool
    """

    return MaestralLock(config_name).locked()


def _wait_for_startup(config_name, timeout=8):
    """Checks if we can communicate with the maestral daemon. Returns ``Start.Ok`` if
    communication succeeds within timeout, ``Start.Failed``  otherwise."""

    sock_name = sockpath_for_config(config_name)
    maestral_daemon = Proxy(URI.format(_escape_spaces(config_name), './u:' + sock_name))

    while timeout > 0:
        try:
            maestral_daemon._pyroBind()
            return Start.Ok
        except Exception:
            time.sleep(0.2)
            timeout -= 0.2
        finally:
            maestral_daemon._pyroRelease()

    return Start.Failed


# ==== main functions to manage daemon ===================================================

def start_maestral_daemon(config_name='maestral', log_to_stdout=False):
    """
    Starts the Maestral daemon with event loop in the current thread. Startup is race
    free: there will never be two daemons running for the same config.

    Wraps :class:`main.Maestral` as Pyro daemon object, creates a new instance and starts
    Pyro's event loop to listen for requests on a unix domain socket. This call will block
    until the event loop shuts down.

    :param str config_name: The name of the Maestral configuration to use.
    :param bool log_to_stdout: If ``True``, write logs to stdout. Defaults to ``False``.
    :raises: :class:`RuntimeError` if a daemon for the given ``config_name`` is already
        running.
    """
    import threading
    from maestral.main import Maestral

    sockpath = sockpath_for_config(config_name)

    if threading.current_thread() is threading.main_thread():
        signal.signal(signal.SIGTERM, _sigterm_handler)

    # acquire PID lock file
    lock = MaestralLock(config_name)
    if not lock.acquire():
        raise RuntimeError('Maestral daemon is already running')

    # Nice ourselves to give other processes priority. We will likely only
    # have significant CPU usage in case of many concurrent downloads.
    os.nice(10)

    try:
        # clean up old socket
        try:
            os.remove(sockpath)
        except FileNotFoundError:
            pass

        # expose maestral as Pyro server
        # convert selected methods to one way calls so that they don't block
        ExposedMaestral = expose(Maestral)

        ExposedMaestral.start_sync = oneway(ExposedMaestral.start_sync)
        ExposedMaestral.stop_sync = oneway(ExposedMaestral.stop_sync)
        ExposedMaestral.pause_sync = oneway(ExposedMaestral.pause_sync)
        ExposedMaestral.resume_sync = oneway(ExposedMaestral.resume_sync)
        ExposedMaestral.shutdown_pyro_daemon = oneway(ExposedMaestral.shutdown_pyro_daemon)

        m = ExposedMaestral(config_name, log_to_stdout=log_to_stdout)

        with Daemon(unixsocket=sockpath) as daemon:
            daemon.register(m, f'maestral.{_escape_spaces(config_name)}')
            daemon.requestLoop(loopCondition=m._loop_condition)

    except Exception:
        traceback.print_exc()
    except (KeyboardInterrupt, SystemExit):
        sys.exit(0)
    finally:
        lock.release()


def start_maestral_daemon_thread(config_name='maestral', log_to_stdout=False):
    """
    Starts the Maestral daemon in a new thread by calling :func:`start_maestral_daemon`.
    Startup is race free: there will never be two daemons running for the same config.

    :param str config_name: The name of the Maestral configuration to use.
    :param bool log_to_stdout: If ``True``, write logs to stdout. Defaults to ``False``.
    :returns: ``Start.Ok`` if successful, ``Start.AlreadyRunning`` if the daemon was
        already running or ``Start.Failed`` if startup failed. It is possible that
        Start.Ok is returned instead of Start.AlreadyRunning in case of a race.
    """

    if is_running(config_name):
        return Start.AlreadyRunning

    t = threading.Thread(
        target=start_maestral_daemon,
        args=(config_name, log_to_stdout),
        name=f'maestral-daemon-{config_name}',
        daemon=True,
    )
    t.start()

    threads[config_name] = t

    if threading.current_thread() is threading.main_thread():
        signal.signal(signal.SIGTERM, _sigterm_handler)

    return _wait_for_startup(config_name)


def _subprocess_launcher(config_name, log_to_stdout):

    if IS_FROZEN:
        subprocess.Popen([sys.executable, '--frozen-daemon', '-c', config_name],
                         start_new_session=True)
    else:
        cc = quote(config_name).strip("'")  # protect against injection
        std_log = bool(log_to_stdout)

        cmd = (f'import maestral.daemon; '
               f'maestral.daemon.start_maestral_daemon("{cc}", {std_log})')

        subprocess.Popen([sys.executable, '-c', cmd], start_new_session=True)


def start_maestral_daemon_process(config_name='maestral', log_to_stdout=False, detach=True):
    """
    Starts the Maestral daemon in a new process by calling :func:`start_maestral_daemon`.
    Startup is race free: there will never be two daemons running for the same config.

    This function assumes that ``sys.executable`` points to the Python executable or a
    frozen executable. In case of a frozen executable, the executable must take the
    command line argument '--frozen-daemon' to start a daemon process which is *not
    syncing* by calling :func:`start_maestral_daemon`. This is currently supported through
    :func:`freeze_support` which should be called from the main entry point, as soon as
    possible after startup.

    :param str config_name: The name of the Maestral configuration to use.
    :param bool log_to_stdout: If ``True``, write logs to stdout. Defaults to ``False``.
    :param bool detach: If ``True``, the daemon process will be detached by double-forking.
    :returns: ``Start.Ok`` if successful, ``Start.AlreadyRunning`` if the daemon was
        already running or ``Start.Failed`` if startup failed. It is possible that
        Start.Ok is returned instead of Start.AlreadyRunning in case of a race.
    """

    if is_running(config_name):
        return Start.AlreadyRunning

    if detach:
        _subprocess_launcher(config_name, log_to_stdout)

    else:
        import multiprocessing as mp
        ctx = mp.get_context('spawn' if IS_MACOS else 'fork')

        ctx.Process(
            target=start_maestral_daemon,
            args=(config_name, log_to_stdout),
            name='maestral-daemon',
            daemon=True,
        ).start()

    return _wait_for_startup(config_name)


def stop_maestral_daemon_process(config_name='maestral', timeout=10):
    """Stops a maestral daemon process by finding its PID and shutting it down.

    This function first tries to shut down Maestral gracefully. If this fails and we know
    its PID, it will send SIGTERM. If that fails as well, it will send SIGKILL to the
    process.

    :param str config_name: The name of the Maestral configuration to use.
    :param float timeout: Number of sec to wait for daemon to shut down before killing it.
    :returns: ``Exit.Ok`` if successful, ``Exit.Killed`` if killed, ``Exit.NotRunning`` if
        the daemon was not running and ``Exit.Failed`` if killing the process failed
        because we could not retrieve its PID.
    """

    if not is_running(config_name):
        return Exit.NotRunning

    pid = get_maestral_pid(config_name)

    try:
        with MaestralProxy(config_name) as m:
            m.stop_sync()
            m.shutdown_pyro_daemon()
    except Pyro5.errors.CommunicationError:
        if pid:
            _send_term(pid)
    finally:
        while timeout > 0:
            if not is_running(config_name):
                return Exit.Ok
            else:
                time.sleep(0.2)
                timeout -= 0.2

        # send SIGTERM after timeout and delete PID file
        _send_term(pid)

        time.sleep(1)

        if not is_running(config_name):
            return Exit.Ok
        elif pid:
            os.kill(pid, signal.SIGKILL)
            return Exit.Killed
        else:
            return Exit.Failed


def stop_maestral_daemon_thread(config_name='maestral', timeout=10):
    """Stops a maestral daemon thread without killing the parent process.

    :param str config_name: The name of the Maestral configuration to use.
    :param float timeout: Number of sec to wait for daemon to shut down before killing it.
    :returns: ``Exit.Ok`` if successful,``Exit.NotRunning`` if the daemon was not running,
        ``Exit.Failed`` if it could not be stopped within timeout.
    """

    if not is_running(config_name):
        return Exit.NotRunning

    # tell maestral daemon to shut down
    try:
        with MaestralProxy(config_name) as m:
            m.stop_sync()
            m.shutdown_pyro_daemon()
    except Pyro5.errors.CommunicationError:
        return Exit.Failed

    # wait for maestral to carry out shutdown
    t = threads.get(config_name)
    t.join(timeout=timeout)
    if t.is_alive():
        return Exit.Failed
    else:
        return Exit.Ok


def get_maestral_proxy(config_name='maestral', fallback=False):
    """
    Returns a Pyro proxy of the a running Maestral instance.

    :param str config_name: The name of the Maestral configuration to use.
    :param bool fallback: If ``True``, a new instance of Maestral will be returned when
        the daemon cannot be reached. Defaults to ``False``.
    :returns: Pyro proxy of Maestral or a new instance.
    :raises: :class:`Pyro5.errors.CommunicationError` if the daemon cannot be reached and
        ``fallback`` is ``False``.
    """

    if is_running(config_name):
        sock_name = sockpath_for_config(config_name)

        sys.excepthook = Pyro5.errors.excepthook
        maestral_daemon = Proxy(URI.format(_escape_spaces(config_name), './u:' + sock_name))
        try:
            maestral_daemon._pyroBind()
            return maestral_daemon
        except Pyro5.errors.CommunicationError:
            maestral_daemon._pyroRelease()

    if fallback:
        from maestral.main import Maestral
        return Maestral(config_name)
    else:
        raise Pyro5.errors.CommunicationError


class MaestralProxy:
    """A context manager to open and close a proxy to the Maestral daemon."""

    def __init__(self, config_name='maestral', fallback=False):
        self.m = get_maestral_proxy(config_name, fallback)

    def __enter__(self):
        return self.m

    def __exit__(self, exc_type, exc_value, tb):
        if isinstance(self.m, Proxy):
            self.m._pyroRelease()

        del self.m



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
# Copyright 2012 Google, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watchdog.observers.fsevents import (
    FSEventsEmitter, FSEventsObserver, DirectorySnapshot,
    FileDeletedEvent, FileModifiedEvent, FileMovedEvent, FileCreatedEvent,
    DirDeletedEvent, DirModifiedEvent, DirMovedEvent, DirCreatedEvent,
    DEFAULT_OBSERVER_TIMEOUT, BaseObserver
)


class OrderedFSEventsEmitter(FSEventsEmitter):
    """
    This subclasses FSEventsEmitter to guarantee an order of events which can be applied
    to reproduce the new state from the old state.

    Looking at the source code for DirectorySnapshotDiff, the event types are categorised
    as follows:

    Created event: The inode is unique to the new snapshot. The path may be unique to the
    new snapshot or exist in both. In the second case, there will be a preceding Deleted
    event or a Moved event with the path as starting point (the old item was deleted or
    moved away).

    Deleted event: The inode is unique to the old snapshot. The path may be unique to the
    old snapshot or exist in both. In the second case, there will be a subsequent Created
    event or a Moved event with the path as end point (something else was created at or
    moved to the location).

    Moved event: The inode exists in both snapshots but with different paths.

    Modified event: The inode exists in both snapshots and the mtime or file size are
    different. DirectorySnapshotDiff will always use the inode’s path from the old
    snapshot.

    From the above classification, there can be at most two created/deleted/moved events
    that share the same path in one snapshot diff:

        Deleted(path1)      + Created(path1)
        Moved(path1, path2) + Created(path1)
        Deleted(path1)      + Moved(path0, path1)

    And any Modified event will come before a Moved event or stand alone. Modified events
    will never be combined by themselves with created or deleted events because they
    require the inode to be present in both snapshots.

    From the above, we could achieve correct ordering for every path by always adding
    Deleted events to the queue first, Modified events second, Moved events third and
    Created events last:

        Deleted -> Modified -> Moved -> Created

    The ordering won’t be correct between unrelated paths and between files and folder.
    The first does not matter for syncing. We solve the second by assuming that when a
    directory is deleted, so are its children. And before a child is created, there must
    be a directory.

    MovedEvents which are not unique (their paths appear in other events) will be split
    into Deleted and Created events by Maestral.
    """

    def queue_events(self, timeout):
        with self._lock:
            if not self.watch.is_recursive and self.watch.path not in self.pathnames:
                return
            new_snapshot = DirectorySnapshot(self.watch.path, self.watch.is_recursive)
            events = new_snapshot - self.snapshot
            self.snapshot = new_snapshot

            # Files.
            for src_path in events.files_deleted:
                self.queue_event(FileDeletedEvent(src_path))
            for src_path in events.files_modified:
                self.queue_event(FileModifiedEvent(src_path))
            for src_path, dest_path in events.files_moved:
                self.queue_event(FileMovedEvent(src_path, dest_path))
            for src_path in events.files_created:
                self.queue_event(FileCreatedEvent(src_path))

            # Directories.
            for src_path in events.dirs_deleted:
                self.queue_event(DirDeletedEvent(src_path))
            for src_path in events.dirs_modified:
                self.queue_event(DirModifiedEvent(src_path))
            for src_path, dest_path in events.dirs_moved:
                self.queue_event(DirMovedEvent(src_path, dest_path))
            for src_path in events.dirs_created:
                self.queue_event(DirCreatedEvent(src_path))


class OrderedFSEventsObserver(FSEventsObserver):

    def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
        BaseObserver.__init__(self, emitter_class=OrderedFSEventsEmitter, timeout=timeout)



# -*- coding: utf-8 -*-
"""
@author: Sam Schott  (ss2151@cam.ac.uk)

(c) Sam Schott; This work is licensed under the MIT licence.

This module contains functions to retrieve platform dependent locations to store app data.
It supports macOS and Linux.

"""

# system imports
import os
import os.path as osp
import platform
import tempfile

# local imports
from maestral.config.base import get_home_dir, get_conf_path, get_data_path, _to_full_path


__all__ = [
    'get_home_dir', 'get_conf_path', 'get_data_path', 'get_log_path',
    'get_cache_path', 'get_autostart_path', 'get_runtime_path'
]

_home_dir = get_home_dir()


def get_cache_path(subfolder=None, filename=None, create=True):
    """
    Returns the default cache path for the platform. This will be:

        - macOS: '~/Library/Application Support/SUBFOLDER/FILENAME'
        - Linux: '$XDG_CACHE_HOME/SUBFOLDER/FILENAME'
        - fallback: '$HOME/.cache/SUBFOLDER/FILENAME'

    :param Optional[str] subfolder: The subfolder for the app.
    :param Optional[str] filename: The filename to append for the app.
    :param bool create: If ``True``, the folder '<subfolder>' will be created on-demand.
    """
    if platform.system() == 'Darwin':
        cache_path = get_conf_path(create=False)
    elif platform.system() == 'Linux':
        fallback = osp.join(_home_dir, '.cache')
        cache_path = os.environ.get('XDG_CACHE_HOME', fallback)
    else:
        raise RuntimeError('Platform not supported')

    return _to_full_path(cache_path, subfolder, filename, create)


def get_log_path(subfolder=None, filename=None, create=True):
    """
    Returns the default log path for the platform. This will be:

        - macOS: '~/Library/Logs/SUBFOLDER/FILENAME'
        - Linux: '$XDG_CACHE_HOME/SUBFOLDER/FILENAME'
        - fallback: '$HOME/.cache/SUBFOLDER/FILENAME'

    :param Optional[str] subfolder: The subfolder for the app.
    :param Optional[str] filename: The filename to append for the app.
    :param bool create: If ``True``, the folder '<subfolder>' will be created on-demand.
    """

    if platform.system() == 'Darwin':
        log_path = osp.join(_home_dir, 'Library', 'Logs')
    elif platform.system() == 'Linux':
        log_path = get_cache_path(create=False)
    else:
        raise RuntimeError('Platform not supported')

    return _to_full_path(log_path, subfolder, filename, create)


def get_autostart_path(filename=None, create=True):
    """
    Returns the default path for login items for the platform. This will be:

        - macOS: '~/Library/LaunchAgents/FILENAME'
        - Linux: '$XDG_CONFIG_HOME/autostart/FILENAME'
        - fallback: '$HOME/.config/autostart/FILENAME'

    :param Optional[str] filename: The filename to append for the app.
    :param bool create: If ``True``, the folder '<subfolder>' will be created on-demand.
    """
    if platform.system() == 'Darwin':
        autostart_path = osp.join(_home_dir, 'Library', 'LaunchAgents')
    elif platform.system() == 'Linux':
        autostart_path = get_conf_path('autostart', create=create)
    else:
        raise RuntimeError('Platform not supported')

    if filename:
        autostart_path = osp.join(autostart_path, filename)

    return autostart_path


def get_runtime_path(subfolder=None, filename=None, create=True):
    """
    Returns the default runtime path for the platform. This will be:

        - macOS: '~/Library/Application Support/SUBFOLDER/FILENAME'
        - Linux: '$XDG_RUNTIME_DIR/SUBFOLDER/FILENAME'
        - fallback: '$HOME/.cache/SUBFOLDER/FILENAME'

    :param Optional[str] subfolder: The subfolder for the app.
    :param Optional[str] filename: The filename to append for the app.
    :param bool create: If ``True``, the folder '<subfolder>' will be created on-demand.
    """

    if platform.system() == 'Darwin':
        runtime_path = get_conf_path(create=False)
    elif platform.system() == 'Linux':
        fallback = get_cache_path(create=False)
        runtime_path = os.environ.get('XDG_RUNTIME_DIR', fallback)
    else:
        raise RuntimeError('Platform not supported')

    return _to_full_path(runtime_path, subfolder, filename, create)


def get_old_runtime_path(subfolder=None, filename=None, create=True):
    """
    Returns the default runtime path for the platform. This will be:

        - macOS: tempfile.gettempdir() + 'SUBFOLDER/FILENAME'
        - Linux: '$XDG_RUNTIME_DIR/SUBFOLDER/FILENAME'
        - fallback: '$HOME/.cache/SUBFOLDER/FILENAME'

    :param Optional[str] subfolder: The subfolder for the app.
    :param Optional[str] filename: The filename to append for the app.
    :param bool create: If ``True``, the folder '<subfolder>' will be created on-demand.
    """

    if platform.system() == 'Darwin':
        runtime_path = tempfile.gettempdir()
    elif platform.system() == 'Linux':
        fallback = get_cache_path(create=False)
        runtime_path = os.environ.get('XDG_RUNTIME_DIR', fallback)
    else:
        raise RuntimeError('Platform not supported')

    return _to_full_path(runtime_path, subfolder, filename, create)



# -*- coding: utf-8 -*-
"""
@author: Sam Schott  (ss2151@cam.ac.uk)

(c) Sam Schott; This work is licensed under the MIT licence.

"""

# external imports
from jeepney import MessageGenerator, new_method_call


class FreedesktopNotifications(MessageGenerator):
    interface = 'org.freedesktop.Notifications'

    def __init__(self, object_path='/org/freedesktop/Notifications',
                 bus_name='org.freedesktop.Notifications'):
        super().__init__(object_path=object_path, bus_name=bus_name)

    def Notify(self, arg_0, arg_1, arg_2, arg_3, arg_4, arg_5, arg_6, arg_7):
        return new_method_call(self, 'Notify', 'susssasa{sv}i',
                               (arg_0, arg_1, arg_2, arg_3, arg_4, arg_5, arg_6,
                                arg_7))

    def CloseNotification(self, arg_0):
        return new_method_call(self, 'CloseNotification', 'u',
                               (arg_0,))

    def GetCapabilities(self):
        return new_method_call(self, 'GetCapabilities')

    def GetServerInformation(self):
        return new_method_call(self, 'GetServerInformation')



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""

This module provides user configuration file management and is mostly copied from the
config module of the Spyder IDE.

"""

import ast
import os
import os.path as osp
import re
import shutil
import time
import configparser as cp
from threading import RLock
import logging


logger = logging.getLogger(__name__)


# =============================================================================
# Auxiliary classes
# =============================================================================

class NoDefault:
    pass


# =============================================================================
# Defaults class
# =============================================================================

class DefaultsConfig(cp.ConfigParser):
    """
    Class used to save defaults to a file and as base class for UserConfig.
    """

    _lock = RLock()

    def __init__(self, path, name, suffix):
        super(DefaultsConfig, self).__init__(interpolation=None)

        self._path = path
        self._name = name
        self._suffix = suffix

        if not osp.isdir(osp.dirname(self._path)):
            os.makedirs(osp.dirname(self._path))

    def _set(self, section, option, value):
        """Private set method"""
        if not self.has_section(section):
            self.add_section(section)
        if not isinstance(value, str):
            value = repr(value)

        super(DefaultsConfig, self).set(section, option, value)

    def save(self):
        """Save config into the associated file."""
        fpath = self.get_config_fpath()

        # See spyder-ide/spyder#1086 and spyder-ide/spyder#1242 for background
        # on why this method contains all the exception handling.

        with self._lock:
            try:
                # The "easy" way
                self.__write_file(fpath)
            except EnvironmentError:
                try:
                    # The "delete and sleep" way
                    if osp.isfile(fpath):
                        os.remove(fpath)

                    time.sleep(0.05)
                    self.__write_file(fpath)
                except Exception:
                    logger.exception('Failed to write user configuration file to disk')

    def __write_file(self, fpath):
        with open(fpath, 'w', encoding='utf-8') as configfile:
            self.write(configfile)

    def get_config_fpath(self):
        """Return the ini file where this configuration is stored."""
        return osp.join(self._path, self._name + self._suffix)

    def set_defaults(self, defaults):
        """Set default values and save to defaults folder location."""
        for section, options in defaults:
            for option in options:
                new_value = options[option]
                self._set(section, option, new_value)


# =============================================================================
# User config class
# =============================================================================

class UserConfig(DefaultsConfig):
    """
    UserConfig class, based on ConfigParser.

    Parameters
    ----------
    path: str
        Configuration file will be saved to this path.
    defaults: {} or [(str, {}),]
        Dictionary containing options *or* list of tuples (sec_name, options)
    load: bool
        If a previous configuration file is found, load will take the values
        from this existing file, instead of using default values.
    version: str
        version of the configuration file in 'major.minor.micro' format.
    backup: bool
        A backup will be created on version changes and on initial setup.
    raw_mode: bool
        If `True` do not apply any automatic conversion on values read from
        the configuration.
    remove_obsolete: bool
        If `True`, values that were removed from the configuration on version
        change, are removed from the saved configuration file.

    Notes
    -----
    The 'get' and 'set' arguments number and type differ from the reimplemented
    methods. 'defaults' is an attribute and not a method.
    """
    DEFAULT_SECTION_NAME = 'main'

    def __init__(self, path, name, defaults=None, load=True, version=None,
                 backup=False, raw_mode=False, remove_obsolete=False, suffix='.ini'):
        """UserConfig class, based on ConfigParser."""
        super(UserConfig, self).__init__(path=path, name=name, suffix=suffix)

        self._load = load
        self._version = self._check_version(version)
        self._backup = backup
        self._raw = 1 if raw_mode else 0
        self._remove_obsolete = remove_obsolete

        self._defaults_folder = 'defaults'
        self._backup_folder = 'backups'
        self._backup_suffix = '.bak'
        self._defaults_name_prefix = 'defaults'

        # This attribute is overriding a method from cp.ConfigParser
        self.defaults = self._check_defaults(defaults)

        if backup:
            self._make_backup()

        if load:
            # If config file already exists, it overrides Default options
            previous_fpath = self.get_previous_config_fpath()
            self._load_from_ini(previous_fpath)
            old_version = self.get_version(version)
            self._old_version = old_version

            # Save new defaults
            self._save_new_defaults(self.defaults)

            # Updating defaults only if major/minor version is different
            major_ver = self._get_major_version(version)
            major_old_ver = self._get_major_version(self._old_version)

            minor_ver = self._get_minor_version(version)
            minor_old_ver = self._get_minor_version(self._old_version)

            if minor_ver != minor_old_ver:

                if backup:
                    self._make_backup(version=old_version)

                self.apply_configuration_patches(old_version=old_version)

                # Remove deprecated options if major version has changed
                if remove_obsolete and major_ver != major_old_ver:
                    self._remove_deprecated_options(old_version)

                # Set new version number
                self.set_version(version, save=False)

            if defaults is None:
                # If no defaults are defined set file settings as default
                self.set_as_defaults()

    # --- Helpers and checkers -----------------------------------------------------------
    @staticmethod
    def _get_minor_version(version):
        """Return the 'major.minor' components of the version."""
        return version[:version.rfind('.')]

    @staticmethod
    def _get_major_version(version):
        """Return the 'major' component of the version."""
        return version[:version.find('.')]

    @staticmethod
    def _check_version(version):
        """Check version is compliant with format."""
        if version is not None:
            regex_check = re.match(r'^(\d+).(\d+).(\d+)$', version)
            if regex_check is None:
                raise ValueError('Version number {} is incorrect - must be in '
                                 'major.minor.micro format'.format(version))

        return version

    def _check_defaults(self, defaults):
        """Check if defaults are valid and update defaults values."""
        if defaults is None:
            defaults = [(self.DEFAULT_SECTION_NAME, {})]
        elif isinstance(defaults, dict):
            defaults = [(self.DEFAULT_SECTION_NAME, defaults)]
        elif isinstance(defaults, list):
            # Check is a list of tuples with strings and dictionaries
            for sec, options in defaults:
                assert isinstance(sec, str)
                assert isinstance(options, dict)
                for opt, _ in options.items():
                    assert isinstance(opt, str)
        else:
            raise ValueError('`defaults` must be a dict or a list of tuples!')

        # This attribute is overriding a method from cp.ConfigParser
        self.defaults = defaults

        if defaults is not None:
            self.reset_to_defaults(save=False)

        for sec, options in defaults:
            if sec == self.DEFAULT_SECTION_NAME:
                options['version'] = self._version

        return defaults

    @classmethod
    def _check_section_option(cls, section, option):
        """Check section and option types."""
        if section is None:
            section = cls.DEFAULT_SECTION_NAME
        elif not isinstance(section, str):
            raise RuntimeError("Argument 'section' must be a string")

        if not isinstance(option, str):
            raise RuntimeError("Argument 'option' must be a string")

        return section

    def _make_backup(self, version=None, old_version=None):
        """
        Make a backup of the configuration file.

        If `old_version` is `None` a normal backup is made. If `old_version`
        is provided, then the backup was requested for minor version changes
        and appends the version number to the backup file.
        """
        fpath = self.get_config_fpath()
        fpath_backup = self.get_backup_fpath_from_version(
            version=version, old_version=old_version)
        path = os.path.dirname(fpath_backup)

        if not osp.isdir(path):
            os.makedirs(path)

        try:
            shutil.copyfile(fpath, fpath_backup)
        except IOError:
            pass

    def _load_from_ini(self, fpath):
        """Load config from the associated file found at `fpath`."""

        with self._lock:
            try:
                self.read(fpath, encoding='utf-8')
            except cp.MissingSectionHeaderError:
                logger.error('File contains no section headers.')

    def _load_old_defaults(self, old_version):
        """Read old defaults."""
        old_defaults = cp.ConfigParser()
        fpath = self.get_defaults_fpath_from_version(old_version)
        old_defaults.read(fpath)
        return old_defaults

    def _save_new_defaults(self, defaults):
        """Save new defaults."""
        path, name = self.get_defaults_path_name_from_version()
        new_defaults = DefaultsConfig(path=path, name=name, suffix=self._suffix)
        if not osp.isfile(new_defaults.get_config_fpath()):
            new_defaults.set_defaults(defaults)
            new_defaults.save()

    def _update_defaults(self, defaults, old_version):
        """Update defaults after a change in version."""
        old_defaults = self._load_old_defaults(old_version)
        for section, options in defaults:
            for option in options:
                new_value = options[option]
                try:
                    old_val = old_defaults.get(section, option)
                except (cp.NoSectionError, cp.NoOptionError):
                    old_val = None

                if old_val is None or str(new_value) != old_val:
                    self._set(section, option, new_value)

    def _remove_deprecated_options(self, old_version):
        """
        Remove options which are present in the file but not in defaults.
        """
        for section in self.sections():
            for option, _ in self.items(section, raw=self._raw):
                if self.get_default(section, option) is NoDefault:
                    try:
                        self.remove_option(section, option)
                        if len(self.items(section, raw=self._raw)) == 0:
                            self.remove_section(section)
                    except cp.NoSectionError:
                        self.remove_section(section)

    # --- Compatibility API --------------------------------------------------------------

    def get_previous_config_fpath(self):
        """Return the last configuration file used if found."""
        return self.get_config_fpath()

    def get_config_fpath_from_version(self, version=None):
        """
        Return the configuration path for given version.

        If no version is provided, it returns the current file path.
        """
        return self.get_config_fpath()

    def get_backup_fpath_from_version(self, version=None, old_version=None):
        """
        Get backup location based on version.

        `old_version` can be used for checking compatibility whereas `version`
        relates to adding the version to the file name.

        To be reimplemented if versions changed backup location.
        """
        fpath = self.get_config_fpath()
        path = osp.join(osp.dirname(fpath), self._backup_folder)
        new_fpath = osp.join(path, osp.basename(fpath))
        if version is None:
            backup_fpath = '{}{}'.format(new_fpath, self._backup_suffix)
        else:
            backup_fpath = "{}-{}{}".format(new_fpath, version, self._backup_suffix)
        return backup_fpath

    def get_defaults_path_name_from_version(self, old_version=None):
        """
        Get defaults location based on version.

        To be reimplemented if versions changed defaults location.
        """
        version = old_version if old_version else self._version
        defaults_path = osp.join(osp.dirname(self.get_config_fpath()),
                                 self._defaults_folder)
        if version is None:
            name = '{}-{}'.format(self._defaults_name_prefix, self._name)
        else:
            name = '{}-{}-{}'.format(self._defaults_name_prefix, self._name, version)
        if not osp.isdir(defaults_path):
            os.makedirs(defaults_path)

        return defaults_path, name

    def get_defaults_fpath_from_version(self, old_version=None):
        """
        Get defaults location based on version.

        To be reimplemented if versions changed defaults location.
        """
        defaults_path, name = self.get_defaults_path_name_from_version(old_version)

        return osp.join(defaults_path, name + self._suffix)

    def apply_configuration_patches(self, old_version=None):
        """
        Apply any patch to configuration values on version changes.

        To be reimplemented if patches to configuration values are needed.
        """
        pass

    # --- Public API ---------------------------------------------------------------------
    def get_version(self, version='0.0.0'):
        """Return configuration (not application!) version."""
        return self.get(self.DEFAULT_SECTION_NAME, 'version', version)

    def set_version(self, version='0.0.0', save=True):
        """Set configuration (not application!) version."""
        version = self._check_version(version)
        self.set(self.DEFAULT_SECTION_NAME, 'version', version, save=save)

    def reset_to_defaults(self, save=True, section=None):
        """Reset config to Default values."""
        for sec, options in self.defaults:
            if section is None or section == sec:
                for option in options:
                    value = options[option]
                    self._set(sec, option, value)
        if save:
            self.save()

    def set_as_defaults(self):
        """Set defaults from the current config."""
        self.defaults = []
        for section in self.sections():
            secdict = {}
            for option, value in self.items(section, raw=self._raw):
                secdict[option] = value
            self.defaults.append((section, secdict))

    def get_default(self, section, option):
        """
        Get default value for a given `section` and `option`.

        This is useful for type checking in `get` method.
        """
        section = self._check_section_option(section, option)
        for sec, options in self.defaults:
            if sec == section:
                if option in options:
                    value = options[option]
                    break
        else:
            value = NoDefault

        return value

    def get(self, section, option, default=NoDefault):
        """
        Get an option.

        Parameters
        ----------
        section: str
            Section name. If `None` is provide use the default section name.
        option: str
            Option name for `section`.
        default:
            Default value (if not specified, an exception will be raised if
            option doesn't exist).
        """
        section = self._check_section_option(section, option)

        if not self.has_section(section):
            if default is NoDefault:
                raise cp.NoSectionError(section)
            else:
                self.add_section(section)

        if not self.has_option(section, option):
            if default is NoDefault:
                raise cp.NoOptionError(option, section)
            else:
                self.set(section, option, default)
                return default

        value = super(UserConfig, self).get(section, option, raw=self._raw)

        default_value = self.get_default(section, option)
        if isinstance(default_value, bool):
            value = ast.literal_eval(value)
        elif isinstance(default_value, float):
            value = float(value)
        elif isinstance(default_value, int):
            value = int(value)
        else:
            try:
                # Lists, tuples, None, ...
                value = ast.literal_eval(value)
            except (SyntaxError, ValueError):
                pass

        if default_value is not NoDefault and type(default_value) is not type(value):
            logger.error(f'Inconsistent config type for [{section}][{option}]. '
                         f'Expected {default_value.__class__.__name__} but '
                         f'got {value.__class__.__name__}.')

        return value

    def set_default(self, section, option, default_value):
        """
        Set Default value for a given `section`, `option`.

        If no defaults exist, no default is created. To be able to set
        defaults, a call to set_as_defaults is needed to create defaults
        based on current values.
        """
        section = self._check_section_option(section, option)
        for sec, options in self.defaults:
            if sec == section:
                options[option] = default_value

    def set(self, section, option, value, save=True):
        """
        Set an `option` on a given `section`.

        If section is None, the `option` is added to the default section.
        """
        section = self._check_section_option(section, option)
        default_value = self.get_default(section, option)

        if default_value is NoDefault:
            default_value = value
            self.set_default(section, option, default_value)

        if isinstance(default_value, bool):
            value = bool(value)
        elif isinstance(default_value, float):
            value = float(value)
        elif isinstance(default_value, int):
            value = int(value)
        # elif isinstance(default_value, list):
        #     value = list(value)
        # elif isinstance(default_value, tuple):
        #     value = tuple(value)
        elif not isinstance(default_value, str):
            value = repr(value)

        self._set(section, option, value)
        if save:
            self.save()

    def remove_section(self, section):
        """Remove `section` and all options within it."""
        super(UserConfig, self).remove_section(section)
        self.save()

    def remove_option(self, section, option):
        """Remove `option` from `section`."""
        super(UserConfig, self).remove_option(section, option)
        self.save()

    def cleanup(self):
        """Remove files associated with config."""
        fpath = self.get_config_fpath()

        backup_path = osp.join(self._path, self._backup_folder)
        defaults_path = osp.join(self._path, self._defaults_folder)

        # remove config file
        try:
            os.remove(fpath)
        except FileNotFoundError:
            pass

        # remove saved backups
        for file in os.scandir(backup_path):
            if file.name.startswith(self._name):
                try:
                    os.remove(file.path)
                except FileNotFoundError:
                    pass

        # remove saved defaults
        for file in os.scandir(defaults_path):
            if file.name.startswith(f'{self._defaults_name_prefix}-{self._name}'):
                try:
                    os.remove(file.path)
                except FileNotFoundError:
                    pass



# -*- coding: utf-8 -*-
"""
@author: Sam Schott  (ss2151@cam.ac.uk)

(c) Sam Schott; This work is licensed under a Creative Commons
Attribution-NonCommercial-NoDerivs 2.0 UK: England & Wales License.

The following APIs should remain stable for frontends:

* maestral.main.Maestral
* maestral.constants
* maestral.daemon
* maestral.errors
* maestral.utils.appdirs
* maestral.utils.autostart

"""

__version__ = '1.0.3'
__author__ = 'Sam Schott'
__url__ = 'https://github.com/SamSchott/maestral'
